package se.mdh.assist.analysis.wcet;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import org.fordiac.ide.model.libraryElement.Event;
import org.fordiac.ide.model.libraryElement.EventConnection;
import org.fordiac.ide.model.libraryElement.FB;
import org.fordiac.ide.model.libraryElement.InterfaceList;
import org.fordiac.ide.model.libraryElement.SubApp;
import org.fordiac.ide.model.libraryElement.SubAppNetwork;
import se.mdh.assist.analysis.wcet.data.FBWCETData;
import se.mdh.assist.analysis.wcet.data.OutputInfo;
import se.mdh.assist.analysis.wcet.data.PeriodWCETData;
import se.mdh.assist.analysis.wcet.data.WCETAnalysisOptions;
import se.mdh.assist.analysis.wcet.data.WCETData;
import se.mdh.assist.analysis.wcet.data.WCETDataList;

public abstract class NetworkAnalysis extends AbstractWCETAnalysis {

	protected SubAppNetwork network;
	
	public NetworkAnalysis(SubAppNetwork network, WCETAnalysisOptions options) {
		super(options);
		this.network = network;
	}
	
	protected WCETDataList doNetworkAnalysis(Event startEvent) {
		return doNetworkAnalysis(startEvent, new HashSet<Event>());
	}
	
	protected WCETDataList doNetworkAnalysis(Event startEvent, Set<Event> visitedEvents) {
		WCETDataList result = new WCETDataList();
		for (EventConnection connection : startEvent.getOutputConnections()) {
			Event targetEvent = connection.getDestination();
			if (targetEvent == null) {
				System.err.println("Error! No target event for connection '" + connection.getName() + "' from '" + startEvent.getName());
			}
			if (visitedEvents.contains(targetEvent)) {
				// TODO handle loop
				System.out.println("Loop detected at " + startEvent.getName() + " -> " + targetEvent.getName());
				continue;
			}
			
			InterfaceList targetInterface = (InterfaceList) targetEvent.eContainer();

			if (!targetEvent.isIsInput()) {
				OutputInfo newOutputInfo = new OutputInfo();
				newOutputInfo.addEvent(targetEvent.getName());
				result.add(new WCETData(0, newOutputInfo));
				continue;
			}
			
			// pick up the WCET data for the target
			FB targetFB = (FB) targetInterface.eContainer();
			FBWCETData fbWcetData = Helpers.getFBWCETDataForFB(targetFB, options);
			
			if (fbWcetData == null) {
				System.err.println("NO wcet data for " + targetFB.getFBType().getName());
				continue;
			}
			
			WCETDataList wcetDataList = fbWcetData.getEventData(targetEvent.getName());
			
			if (wcetDataList == null) {
				System.err.println("NO wcet data for event input "+targetEvent.getName()+" of " + targetFB.getFBType().getName());
				continue;
			}
			
			// check all WCET data for the target FB
			WCETDataList resultForFB = new WCETDataList();
			visitedEvents.add(targetEvent);
			for (WCETData targetData : wcetDataList) {
				// collect info for all output events, add it together...
				WCETDataList collectedDataList = collectRecursiveWCETData(targetData, targetInterface, visitedEvents);
				
				// if still no collected data, add empty...
				if (collectedDataList.size() == 0) {
					collectedDataList.add(new WCETData());
				}
				// add wcet for currrrent targetData
				for (WCETData data : collectedDataList) {
					data.add(targetData.getWCETValue(), null);
				}
				resultForFB.addAllWithNormalization(collectedDataList, options.normalizationMethod);
			}
			visitedEvents.remove(targetEvent);
			result = WCETDataList.getAllCombinations(result, resultForFB);
			
		}
		
		return result;
	}
	
	protected ArrayList<PeriodWCETData> doPeriodAnalysis() {
		ArrayList<PeriodWCETData> result = new ArrayList<PeriodWCETData>();
		for (FB componentFB : network.getFBs()) {
			InterfaceList targetInterface = componentFB.getInterface();
			FBWCETData fbWCETData = Helpers.getFBWCETDataForFB(componentFB, options);
			if (fbWCETData == null) {
				System.err.println("No WCET data for "+ componentFB.getName()+ ", Type: " + componentFB.getFBType().getName()+"!");
				continue;
			}
			
			for (PeriodWCETData periodWCETData : fbWCETData.getPeriodData()) {
				WCETDataList resultForPeriod = new WCETDataList();
				for (WCETData wcetData : periodWCETData.getData()) {
					WCETDataList collectedDataList = collectRecursiveWCETData(wcetData, targetInterface, new HashSet<Event>());
					
					// if still no collected data, add empty...
					if (collectedDataList.size() == 0) {
						collectedDataList.add(new WCETData());
					}
					// add wcet for currrrent targetData
					for (WCETData data : collectedDataList) {
						data.add(wcetData.getWCETValue(), null);
					}
					resultForPeriod.addAllWithNormalization(collectedDataList, options.normalizationMethod);
				}
				PeriodWCETData newData = new PeriodWCETData(periodWCETData.getPeriod(), resultForPeriod, periodWCETData.getOriginStack());
				newData.addToOriginStack(componentFB.getFBType().getName(), componentFB.getName());
				result.add(newData);
			}
		}
		result.addAll(doSubAppPeriodAnalysis());
		return result;
	}
	
	private Collection<PeriodWCETData> doSubAppPeriodAnalysis() {
		ArrayList<PeriodWCETData> result = new ArrayList<PeriodWCETData>();
		
		for (SubApp subApp : network.getSubApps()) {
			SubAppPeriodAnalysis analysis = new SubAppPeriodAnalysis(subApp.getSubAppNetwork(), options);
			result.addAll(analysis.doPeriodAnalysis());
		}
		return result;
	}

	protected WCETDataList collectRecursiveWCETData(WCETData sourceData, InterfaceList targetInterface, Set<Event> visitedEvents) {
		WCETDataList collectedDataList = new WCETDataList();
		
		Map<String, Integer> outputMap = sourceData.getOutputInfo().getMap();
		for (String outputEventName : outputMap.keySet()) {
			for (Event outputEvent : targetInterface.getEventOutputs()) {
				if(outputEventName.equals(outputEvent.getName())) {
					// get recursive results
					// TODO result caching? make deep clones?
					System.out.println(outputEventName);
					WCETDataList recursiveResults  = doNetworkAnalysis(outputEvent, visitedEvents);
					recursiveResults.multiplyAll(outputMap.get(outputEventName));
					collectedDataList = WCETDataList.getAllCombinations(collectedDataList, recursiveResults);
				}
			}
		}
		return collectedDataList;
	}

	public SubAppNetwork getNetwork() {
		return network;
	}

	public void setNetwork(SubAppNetwork network) {
		this.network = network;
	}
	
	static class SubAppPeriodAnalysis extends NetworkAnalysis {

		public SubAppPeriodAnalysis(SubAppNetwork network,
				WCETAnalysisOptions options) {
			super(network, options);
		}

		@Override
		public FBWCETData doAnalysis() {
			HashMap<String, WCETDataList> eventData = new HashMap<String, WCETDataList>();
			ArrayList<PeriodWCETData> periodData = doPeriodAnalysis();
			return new FBWCETData(eventData, periodData);
		}	
	}

}
